//---------------------------------------------------------------------------
// game.fs  XNA 2.0 game type implementation for Lunar Lander
//
// 2008 written by Phillip Trelford
//---------------------------------------------------------------------------

#light

namespace MicrosoftResearch.LunarLander

    open Font                               
    open System
    open System.Collections.Generic
    open Microsoft.Xna.Framework
    open Microsoft.Xna.Framework.Content
    open Microsoft.Xna.Framework.Graphics
    open Microsoft.Xna.Framework.Input
    open MicrosoftResearch.LunarLander
    //open Microsoft.FSharp.Math.SI

    /// Seconds helper
    module Seconds =
        // Converts TimeSpan to seconds value
        let OfTimespan (ts:TimeSpan) = float32 ts.TotalSeconds * 1.0f<s>

    /// Game states
    type GameState = | Play | GameOver 

    /// This is the main class which runs the game
    type Game() as this =
        inherit Microsoft.Xna.Framework.Game ()

        let mutable graphics = null
        let mutable spriteBatch = null
        let mutable _primitiveBatch = None

        /// Lunar lander craft        
        let lander = new Lander ()
        
        /// Current game state
        let mutable gameState = Play
        /// Landed status
        let mutable landed = false
                
        do
            base.Content.RootDirectory <- "Content"
            graphics <- 
                new GraphicsDeviceManager (this, 
                    PreferredBackBufferHeight = 600, 
                    PreferredBackBufferWidth = 800)
            graphics.ToggleFullScreen ()

        member private this.primitiveBatch = Option.get _primitiveBatch

        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        override this.Initialize() =
            //Audio.Initialize ()
            base.Initialize()

        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        override this.LoadContent () = 
            spriteBatch <- new SpriteBatch (base.GraphicsDevice)
            _primitiveBatch <- Some (new PrimitiveBatch (base.GraphicsDevice))
             
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        override this.UnloadContent () = 
            // TODO: Unload any non ContentManager content here
            ()

        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        override this.Update gameTime =
            // Update audio module                                    
            //Audio.Update ()

            /// Elapsed seconds since last update
            let elapsed = Seconds.OfTimespan gameTime.ElapsedGameTime                                                                                                                                                                

            /// Keys state
            let keyState = Keyboard.GetState ()
            /// Player 1 pad
            let pad = GamePad.GetState (PlayerIndex.One)
            // Allows the game to exit
            if pad.Buttons.Back = ButtonState.Pressed  || keyState.IsKeyDown (Keys.Escape) then this.Exit ()         
            
            if gameState = GameOver then
                if pad.Buttons.Start = ButtonState.Pressed || keyState.IsKeyDown (Keys.Enter) then 
                    lander.Reset ()
                    landed <- false
                    gameState <- Play
            
            /// Left thruster                  
            let leftThrust = 
                match keyState.IsKeyDown (Keys.Z) with
                | true -> 0.25f
                | false -> pad.Triggers.Left                 
            /// Right thruster                    
            let rightThrust = 
                match  keyState.IsKeyDown (Keys.X) with 
                | true -> 0.25f
                | false -> pad.Triggers.Right
            /// Forward thruster                
            let forwardThrust =
                if gameState = Play then                
                    if lander.Fuel > 0.0f<l> then
                        (pad.Buttons.A = ButtonState.Pressed) || keyState.IsKeyDown (Keys.Space)
                    else
                        false                                                                      
                else false
            //Audio.PlayThrust forwardThrust
            if forwardThrust then lander.Fuel <- lander.Fuel - 3.0f<l>
            if lander.Fuel < 0.0f<l> then lander.Fuel <- 0.0f<l>                    
            
            // Update lander with elapsed time and user inputs
            if not (gameState = GameOver) then
                lander.Update elapsed (leftThrust, rightThrust, forwardThrust)
                       
            // Handle ship leaving viewport by teleporting to other side                                    
            let viewPort = this.GraphicsDevice.Viewport            
            if (lander.X < (1.0f<m> * float32 viewPort.X)) then 
                lander.Teleport (lander.X + 1.0f<m> * (float32 viewPort.Width), lander.Y)    
            if (lander.X > (1.0f<m> * float32 (viewPort.X + viewPort.Width) )) then 
                lander.Teleport (lander.X - 1.0f<m> * (float32 viewPort.Width), lander.Y)    
            if (lander.Y < (1.0f<m> * float32 viewPort.Y)) then 
                lander.Teleport (lander.X, lander.Y  + 1.0f<m> * (float32 viewPort.Height))    
            if (lander.Y > (1.0f<m> * float32 (viewPort.Y + viewPort.Height) )) then 
                lander.Teleport (lander.X, lander.Y - 1.0f<m> * (float32 viewPort.Height) )    
                
            // Collision detection between lander and terrain                                                       
            let GetCollisionLine () =                                                                  
                /// Computes lines for a set of vertices 
                let ComputeLines (vertices:Vector2 array) =
                    let length = Array.length vertices 
                    Array.zip (Array.sub vertices 0 (length - 1)) (Array.sub vertices 1 (length-1))
                    |> Array.map (fun (a,b) -> Line2 (a,b))
                                            
                let terrainLines = Terrain.ScaleVertices (viewPort.Width,viewPort.Height) |> ComputeLines
                let landerLines = 
                    lander.MapVertices 
                        (new Vector2 (lander.X/1.0f<m>, lander.Y/1.0f<m>)) 
                    |> ComputeLines
                terrainLines |> Array.tryFind (fun a ->
                    match 
                        landerLines 
                        |> Array.tryFind 
                        (fun b -> Line2.Intersect (a, b) ) with
                    | Some (line) -> true
                    | None -> false
                )
            if gameState = Play then                
                match GetCollisionLine () with
                | Some (line) ->
                    if line.V1.Y = line.V2.Y then 
                        gameState <- GameOver
                        landed <- true
                    else                         
                        //Audio.PlayExplosion ()
                        gameState <- GameOver                        
                | None -> ()                    
            base.Update (gameTime)
            
        /// This is called when the game should draw itself.
        override this.Draw gameTime =
            // Clears screen 
            graphics.GraphicsDevice.Clear (Color.Black)
            
            /// Renders terrain
            let RenderTerrain () = 
                /// View port
                let viewPort = this.GraphicsDevice.Viewport
                /// Screen vertices
                let screenVertices = 
                    Terrain.ScaleVertices 
                        (viewPort.Width,viewPort.Height) 
                    |> Array.map (fun v -> 
                        new Vector2 (v.X / 1.0f, v.Y / 1.0f), Color.White)                                    
                this.primitiveBatch.Render (PrimitiveType.LineStrip, screenVertices)
            RenderTerrain ()                
            
            /// Renders lander            
            let RenderLander () = 
                /// Screen vertices
                let screenVertices = 
                    lander.MapVertices 
                        (new Vector2 (lander.X/1.0f<m>, lander.Y/1.0f<m>)) 
                    |> Array.map (fun v -> v, Color.White)                                            
                this.primitiveBatch.Render (PrimitiveType.LineList, screenVertices)
            //if gameState = Play then 
            RenderLander ()                
                             
            // Renders fuel display
            let fuel = int (lander.Fuel / 1.0f<l>)
            WriteText this.primitiveBatch (10.0f,10.0f) 1.5f Color.White (sprintf "FUEL %05dL" fuel) 
     
            // Renders velocity display
            let dx, dy = lander.Velocity
            let d = sqrt ((dx * dx) + (dy * dy)) / 1.0f<m/s>
            WriteText this.primitiveBatch (400.0f,10.0f) 1.5f Color.White (sprintf "SPEED %0.2fm/s" d) 
                  
            // Show "Game Over" or "Completed" message            
            if gameState = GameOver then 
                let text = if landed then "COMPLETED" else "GAME OVER"
                WriteText this.primitiveBatch (120.0f,200.0f) 4.0f Color.White text
                                                    
            base.Draw (gameTime)
           