//---------------------------------------------------------------------------
// lander.fs    Contains Lunar lander type
//
// 2008 written by Phillip Trelford
//---------------------------------------------------------------------------

#light

namespace MicrosoftResearch.LunarLander

open Microsoft.Xna.Framework
open Microsoft.FSharp.Math
//open SI

[<Measure>] type m
[<Measure>] type s

/// Gravity helper 
module Gravity = 
    let Earth = 9.8f<m/s^2> // Standard Gravity
    let Lunar = Earth / 6.0f

[<Measure>] 
type radians
[<Measure>] 
type rad = radians
[<Measure>] 
type degrees =
    static member to_radians d =
        MathHelper.ToRadians (d / 1.0f<degrees>) * 1.0f<rad>                
[<Measure>] 
type deg = degrees
[<Measure>] 
type litres
[<Measure>]
type l = litres
                                                        
/// Radians helper        
module Radians =           
    let OfDegrees d = 
        MathHelper.ToRadians (d / 1.0f<deg>)
    let Rotate (a:float32<rad>) (x:float32,y:float32) =                               
        let r = a / 1.0f<rad>           
        x * cos(r) - y * sin(r), y * cos(r) + x * sin(r)                     
                
/// Lunar lander type
type Lander () =
    let mutable x = 100.0f<m>
    let mutable y = 200.0f<m>
    let mutable velocity = (0.0f<m/s>,0.0f<m/s>)       
    let mutable angle = degrees.to_radians 0.0f<deg>                                                 
    let mutable angularVelocity = 0.0f<rad/s>
    let mutable fuel = 999.0f<l>
                             
    /// Sets default values            
    let SetDefaults () =
        x <- 100.0f<m>
        y <- 200.0f<m>          
        velocity <- (0.0f<m/s>,0.0f<m/s>) 
        angle <- degrees.to_radians 0.0f<deg>
        angularVelocity <- 0.0f<rad/s>
        fuel <- 999.0f<l>
        
    do SetDefaults ()            
            
    member this.X = x
    member this.Y = y
    member this.Angle = angle
    member this.Velocity = velocity
    member this.Fuel with get () = fuel and set value = fuel <- value
    
    /// Resets values
    member this.Reset () =
        SetDefaults ()
    
    /// Updates lander
    member this.Teleport (newX,newY) =
        x <- newX
        y <- newY
    
    /// Updates lander
    member this.Update elapsed (leftThrust, rightThrust, forwardThrust) = 
        // Capture ship forward acceleration
        let acceleration =
            match forwardThrust with
            | true -> 50.0f<m/s^2>
            | false -> 0.0f<m/s^2>
                                                                                                                 
        // Update ship angle 
        let rotationalThrust = leftThrust - rightThrust                                                                                        
        let rotation = degrees.to_radians 0.05f<deg> * rotationalThrust
        if rotation <> 0.0f<rad> then
            let acceleration = (rotation / elapsed)            
            angularVelocity <- angularVelocity + acceleration                       
        angle <- angle + (angularVelocity * elapsed)
        
        // Update ship velocity
        let heading = (angle - degrees.to_radians 90.0f<deg>) / 1.0f<rad>
        let ddx, ddy = 
            (acceleration * cos (heading)), (acceleration * sin (heading))
        let newVelocity = 
            let dx, dy = velocity 
            (dx + (ddx * elapsed), 
                dy + (ddy * elapsed) + (Gravity.Lunar * elapsed))
        velocity <- newVelocity
        
        // Update ship position
        x <- x + (fst velocity * elapsed)                
        y <- y + (snd velocity * elapsed)           

    /// Map vertices to specified location        
    member this.MapVertices where =
        // how big is the ship?
        let sizeX = 10.0f
        let sizeY = 15.0f
        let cutoutSize = 5.0f                                                                     
        (*[|
            // from the nose, down the left hand side
            (0.0f, -sizeY);
            (-sizeX, sizeY);
            // to the right and up, into the cutout
            (-sizeX, sizeY);
            (0.0f, sizeY - cutoutSize);
            // to the right and down, out of the cutout
            (0.0f, sizeY - cutoutSize);
            (sizeX, sizeY);
            // and back up to the nose, where we started.
            (sizeX, sizeY);
            (0.0f, -sizeY);
        |]*)
        [|
            (-4,-4); (-2,-4);
            (-3,-4); (-3,-3);
            (-3,-3); (-2,-2);
            (-2,-2); (-2, 0);
            (-2,-2); ( 2,-2);
            (-2, 0); ( 2, 0);
            ( 2,-2); ( 2, 0);
            (-1, 0); (-2, 1);
            (-2, 1); (-2, 2);
            (-2, 2); (-1, 3); 
            (-1, 3); ( 1, 3);
            ( 1, 3); ( 2, 2);
            ( 2, 2); ( 2, 1);
            ( 2, 1); ( 1, 0);
            ( 1, 0); (-1, 0);            
            ( 2,-2); ( 3,-3);
            ( 3,-3); ( 3,-4);
            ( 2,-4); ( 4,-4);
            (-2,-3); ( 2,-3);
            (-2,-3); (-1,-2);
            ( 2,-3); ( 1,-2);
        |]
        |> Array.map (fun (x,y) -> x * 4, y * 4)
        |> Array.map (fun (x,y) -> float32 x, float32 y)
        |> Array.map (fun (x,y) -> 
            Radians.Rotate (this.Angle + degrees.to_radians 180.0f<deg>)  (x,y) ) 
        |> Array.map (fun (x,y) -> where + new Vector2 (x,y))
        