//---------------------------------------------------------------------------
// primitivebatch.fs    Primitive batch render of points, lines and triangles
//
// 2008 written by Phillip Trelford
//---------------------------------------------------------------------------

#light

namespace MicrosoftResearch.LunarLander

    open System
    open System.Collections.Generic

    open Microsoft.Xna.Framework;
    open Microsoft.Xna.Framework.Graphics;

    type PrimitiveBatch (graphicsDevice:GraphicsDevice) =
        let mutable _primitiveType = PrimitiveType.PointList
        let mutable hasBegun = false
        let vertices = List<VertexPositionColor>()       
        let vertexDeclaration = new VertexDeclaration(graphicsDevice,VertexPositionColor.VertexElements)
        let basicEffect = new BasicEffect(graphicsDevice, null, VertexColorEnabled = true)
        do  let width, height = graphicsDevice.Viewport.Width, graphicsDevice.Viewport.Height
            basicEffect.Projection <- Matrix.CreateOrthographicOffCenter (0.0f, width |> float32, height |> float32, 0.0f, 0.0f, 1.0f);
                                  
        member this.Begin (primitiveType:PrimitiveType) =
            if hasBegun then raise (new InvalidOperationException("End must be called before Begin can be called again."))
            _primitiveType <- primitiveType          
            vertices.Clear ()
            graphicsDevice.VertexDeclaration <- vertexDeclaration
            basicEffect.Begin();
            basicEffect.CurrentTechnique.Passes.Item(0).Begin ()
            hasBegun <- true;     
            
        member this.AddVertex (vertex:Vector2, color:Color) =
            if not hasBegun then raise (new InvalidOperationException("Begin must be called before AddVertex can be called."))
            vertices.Add (new VertexPositionColor( Vector3(vertex, 0.0f), color))
        
        member this.AddRange (range:(Vector2 * Color) array) =                                
            if not hasBegun then raise (new InvalidOperationException("Begin must be called before AddRange can be called."))
            range |> Array.map (fun (vertex, color) -> new VertexPositionColor( Vector3(vertex, 0.0f), color)) |> vertices.AddRange 
            
        member this.End () =
            if not hasBegun then raise (new InvalidOperationException("Begin must be called before End can be called."))            
            if vertices.Count > 0 then
                let numVerticesPerPrimitive, isList =                 
                    match _primitiveType with 
                    | PrimitiveType.PointList -> 1, true                                     
                    | PrimitiveType.LineList -> 2, true
                    | PrimitiveType.TriangleList -> 3, true
                    | PrimitiveType.LineStrip -> 1, false                    
                    | _ -> raise (new InvalidOperationException("primitive is not valid"))                                                      
                let primitiveCount = vertices.Count / numVerticesPerPrimitive
                if isList then
                    graphicsDevice.DrawUserPrimitives<VertexPositionColor> (_primitiveType, vertices.ToArray (), 0, primitiveCount)                        
                else
                    let indices = Array.init vertices.Count (fun i -> int16 i)
                    graphicsDevice.DrawUserIndexedPrimitives<VertexPositionColor> (_primitiveType, vertices.ToArray (), 0, vertices.Count, indices, 0, vertices.Count - 1)

            basicEffect.CurrentTechnique.Passes.Item(0).End ()
            basicEffect.End ()
            hasBegun <- false
            
        member this.Render (primitiveType:PrimitiveType, range:(Vector2 * Color) array) =
            this.Begin (primitiveType)
            this.AddRange (range)
            this.End ()

